#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <math.h>

#include "asdf/asdf.h"
#include "asdf/cache.h"

#include "tree/eval.h"
#include "tree/packed.h"
#include "util/interval.h"

#include "util/switches.h"
#include "util/macros.h"


// Create a single ASDF node, evaluating the corners without recursion.
ASDF* asdf_root(PackedTree* tree, Region region)
{
    // Allocate an ASDF structure
    ASDF* asdf = calloc(1, sizeof(ASDF));
    
    // Save the corners of the world
    *asdf = (ASDF){
        .state = BRANCH,
        .X = (Interval){region.X[0], region.X[region.ni]},
        .Y = (Interval){region.Y[0], region.Y[region.nj]},
        .Z = (Interval){region.Z[0], region.Z[region.nk]},
    };

    // Sample the distance field at the corners of the cell
    for (int n=0; n < 8; ++n)
        asdf->d[n] = eval_f(tree,
            (n & 4) ? asdf->X.upper : asdf->X.lower,
            (n & 2) ? asdf->Y.upper : asdf->Y.lower,
            (n & 1) ? asdf->Z.upper : asdf->Z.lower);
    
    return asdf;
}

////////////////////////////////////////////////////////////////////////////////

ASDF* build_asdf(PackedTree* tree, Region region, int* halt)
{
    // Special interrupt system, set asynchronously by on high
    if (*halt) return NULL;
    
    // Below a minimum size, do pixel-by-pixel evaluation
    if ((region.ni+1)*(region.nj+1)*(region.nk+1) < MIN_VOLUME)
        return build_asdf_region(tree, region);
    
    // Allocate an ASDF structure
    ASDF* asdf = calloc(1, sizeof(ASDF));
    
    // Save the corners of the world
    *asdf = (ASDF){
        .state = BRANCH,
        .X = (Interval){region.X[0], region.X[region.ni]},
        .Y = (Interval){region.Y[0], region.Y[region.nj]},
        .Z = (Interval){region.Z[0], region.Z[region.nk]},
    };

    // Evaluate across the interval if we've got a volume
    Interval result;
    if (region.voxels > 1) {
        result = eval_i(tree, asdf->X, asdf->Y, asdf->Z);
    } 
    // Otherwise, evaluate the eight corners individually
    else {
        for (int n=0; n < 8; ++n)
            asdf->d[n] = eval_f(tree,
                (n & 4) ? asdf->X.upper : asdf->X.lower,
                (n & 2) ? asdf->Y.upper : asdf->Y.lower,
                (n & 1) ? asdf->Z.upper : asdf->Z.lower);
                
        result = (Interval){asdf->d[0], asdf->d[0]};
        for (int n=0; n < 8; ++n) {
            if (asdf->d[n] < result.lower)  result.lower = asdf->d[n];
            if (asdf->d[n] > result.upper)  result.upper = asdf->d[n];
        }
    }

    // If we get an unambiguous result, mark as filled or empty
    if (result.upper < 0)
        asdf->state = FILLED;
    else if (result.lower >= 0)
        asdf->state = EMPTY;
    else if (region.voxels == 1) {
        asdf->state = LEAF;
        return asdf;
    }

#if PRUNE
    disable_nodes(tree);
#endif
    
    // If the cell is filled or empty, calculate corner values
    // and return without recursing
    if (asdf->state != BRANCH)
        for (int n=0; n < 8; ++n)
            asdf->d[n] = eval_f(tree,
                (n & 4) ? asdf->X.upper : asdf->X.lower,
                (n & 2) ? asdf->Y.upper : asdf->Y.lower,
                (n & 1) ? asdf->Z.upper : asdf->Z.lower);
    
    // Otherwise, split and recurse on the spatial region
    else {
    
        // Fill in up to eight octants, depending on whether we can still
        // split the region along this particular axis
        Region octants[8];
        uint8_t bits = octsect(region, octants);
        for (int i=0; i < 8; ++i)
            if (bits & (1 << i))
                asdf->branches[i] = build_asdf(tree, octants[i], halt);
    
        // If we've halted, then there might be incomplete trees
        // further down the line, so we skip the simplification steps.
        if (!*halt) {
            // Pull d values from children
            get_d_from_children(asdf);
            
            // Merge cells if possible
            simplify(asdf);
        }
    }
    
#if PRUNE
    enable_nodes(tree);
#endif

    return asdf;
}

////////////////////////////////////////////////////////////////////////////////

ASDF* build_asdf_region(PackedTree* tree, Region region)
{  
    int voxels = (region.ni+1)*(region.nj+1)*(region.nk+1);
    
    float *X = malloc(voxels*sizeof(float)),
          *Y = malloc(voxels*sizeof(float)),
          *Z = malloc(voxels*sizeof(float));
    
    // Copy the X, Y, Z vectors into a flattened matrix form.
    int q = 0;
    for (int k = 0; k <= region.nk; ++k)
        for (int j = 0; j <= region.nj; ++j)
            for (int i = 0; i <= region.ni; ++i) {
                X[q] = region.X[i];
                Y[q] = region.Y[j];
                Z[q] = region.Z[k];
                q++;
            }
    
    // Create a dummy region with the flattened vectors
    Region r = (Region) {.X = X, .Y = Y, .Z = Z, .voxels = voxels };
    float* result = eval_r(tree, r);
    
    // Free the allocated matrices
    free(X);
    free(Y);
    free(Z);

    // Calculate k and j stride in the results array
    int kstride = (region.nj+1)*(region.ni+1);
    int jstride = region.ni+1;
    
    // Set these values to zero so that lookups will index
    // into the results array at the correct point
    region.imin = 0;
    region.jmin = 0;
    region.kmin = 0;
    
    return _build_asdf_region(result, region, kstride, jstride);
}

ASDF* _build_asdf_region(float* result, Region region, int kstride, int jstride)
{
    ASDF* asdf = calloc(1, sizeof(ASDF));
    *asdf = (ASDF) {
        .state = BRANCH,
        .X = (Interval){region.X[0], region.X[region.ni]},
        .Y = (Interval){region.Y[0], region.Y[region.nj]},
        .Z = (Interval){region.Z[0], region.Z[region.nk]}
    };
    
    // For a single-cell ASDF, mark as filled or empty based on corners
    if (region.voxels == 1) {
    
        // Populate the corners of the ASDF by looking up values in
        // the array of results.
        int count = 0;
        for (int n=0; n<8; ++n) {
            int i = region.imin + ((n & 4) ? region.ni : 0),
                j = region.jmin + ((n & 2) ? region.nj : 0),
                k = region.kmin + ((n & 1) ? region.nk : 0);
            asdf->d[n] = result[k*kstride + j*jstride + i];
            
            if (asdf->d[n] < 0) count++;
            else                count--;
        }
    
        if (count == 8)         asdf->state = FILLED;
        else if (count == -8)   asdf->state = EMPTY;
        else                    asdf->state = LEAF;
    }
    
    // For a branching ASDF, split and recurse
    else {
        // Split and recurse
        Region octants[8];
        uint8_t bits = octsect(region, octants);
        for (int i=0; i < 8; ++i) {
            if (bits & (1 << i)) {
                asdf->branches[i] = _build_asdf_region(result, octants[i],
                                                       kstride, jstride);
            }
        }
        get_d_from_children(asdf);
        simplify(asdf);
    }
    
    return asdf;
}

////////////////////////////////////////////////////////////////////////////////

void simplify(ASDF* asdf)
{
    // Arrange the axes so that the largest axis is (potentially) merged first
    int axes[3];
    float dx = asdf->X.upper - asdf->X.lower,
          dy = asdf->Y.upper - asdf->Y.lower,
          dz = asdf->Z.upper - asdf->Z.lower;
    if (dx >= dy && dx >= dz) {
        axes[0] = 4;
        axes[1] = dy >= dz ? 2 : 1;
        axes[2] = dy >= dz ? 1 : 2;
    } else if (dy >= dx && dy >= dz) {
        axes[0] = 2;
        axes[1] = dx >= dz ? 4 : 1;
        axes[2] = dx >= dz ? 1 : 4;
    } else {
        axes[0] = 1;
        axes[1] = dx >= dy ? 4 : 2;
        axes[2] = dx >= dy ? 2 : 4;
    }

    // Loop over all three axes
    for (int a=0; a < 3; ++a) {
        int axis = axes[a];
        
        // If the ASDF doesn't split on this axis, skip it.
        if (asdf->branches[axis] == NULL) continue;
        
        // Check to make sure that this axis is mergeable across
        // all (up to) four octant-pairs.
        _Bool merge = true;
        for (int b=0; merge && b < 8; ++b) {
            if (b & axis) continue;
            if (asdf->branches[b] == NULL) continue;
            
            // Pick the octant-pair
            ASDF* A = asdf->branches[b];
            ASDF* B = asdf->branches[b|axis];
            
            // If either are BRANCH, then we can't merge them.
            if (A->state == BRANCH || B->state == BRANCH) {
                merge = false;
                continue;
            }
        
            // If they are both FILLED or EMPTY, we're happy.
            if (A->state != LEAF && A->state == B->state) {
                continue;
            }
        
            // If one of them is a leaf, then we have to check
            // to see if internal values are accurately reconstructed
            // with interpolation
            
            // Pull d values from the pair of cells
            float d[8];
            for (int c=0; c < 8; ++c) {
                d[c] = (c & axis) ? B->d[c] : A->d[c];
            }
            // Pull bounds from the cells
            float xmin = A->X.lower, ymin = A->Y.lower, zmin = A->Z.lower;
            float dx = B->X.upper - xmin,
                  dy = B->Y.upper - ymin,
                  dz = B->Z.upper - zmin;
                
            // Check the d values that lie on the boundary.
            for (int c=0; merge && c < 8; ++c) {
                if (!(c & axis)) continue;
                
                float x = (c & 4) ? A->X.upper : A->X.lower;
                float y = (c & 2) ? A->Y.upper : A->Y.lower;
                float z = (c & 1) ? A->Z.upper : A->Z.lower;
                
                float distance = interpolate(d, x, y, z,
                                      xmin, ymin, zmin,
                                      dx, dy, dz);
    
                // If this corner can't be reconstructed accurately, abort.
                if (fabs(distance - A->d[c]) > EPSILON) {
                    merge = false;
                }
            }
        }
        
        // Move on to the next axis if we can't do this merge
        if (!merge) continue;
            
        // Execute the merge
        for (int b=0; merge && b < 8; ++b) {
            if (b & axis) continue;
            if (asdf->branches[b] == NULL) continue;
            
            merge_cells(asdf->branches[b], asdf->branches[b|axis], axis);
            free_asdf(asdf->branches[b|axis]);
            asdf->branches[b|axis] = NULL;
        }
    }
    
    // If we've merged all of the branches except one, switch the state
    // to something other than BRANCH.
    if (!asdf->branches[1] && !asdf->branches[2] && !asdf->branches[4]) {
        asdf->state = asdf->branches[0]->state;
        free_asdf(asdf->branches[0]);
        asdf->branches[0] = NULL;
    }
}


void merge_cells(ASDF* A, ASDF* B, int axis)
{
    A->X.upper = B->X.upper;
    A->Y.upper = B->Y.upper;
    A->Z.upper = B->Z.upper;
    for (int i=0; i < 8; ++i) {
        if (i & axis) {
            A->d[i] = B->d[i];
        }
    }
    
    if (A->state == LEAF || B->state == LEAF) {
        A->state = LEAF;
    }
}

////////////////////////////////////////////////////////////////////////////////

void asdf_scale(ASDF* asdf, float mm_per_unit)
{
    if (asdf == NULL)   return;
    
    asdf->X.lower *= mm_per_unit;
    asdf->X.upper *= mm_per_unit;
    asdf->Y.lower *= mm_per_unit;
    asdf->Y.upper *= mm_per_unit;
    asdf->Z.lower *= mm_per_unit;
    asdf->Z.upper *= mm_per_unit;
    
    for (int i=0; i < 8; ++i)
        asdf_scale(asdf->branches[i], mm_per_unit);
}

////////////////////////////////////////////////////////////////////////////////

void get_d_from_children(ASDF* asdf)
{
    for (int i=0; i < 8; ++i)
        asdf->d[i] = asdf->branches[0]->d[i];
    
    uint8_t axes = (asdf->branches[1] ? 1 : 0) |
                   (asdf->branches[2] ? 2 : 0) |
                   (asdf->branches[4] ? 4 : 0);

    // Exhaustive case statement defining which cells we should
    // look to for d values, given all possible axis splits.
    switch (axes) {
        case 1: asdf->d[1] = asdf->branches[1]->d[1];
                asdf->d[3] = asdf->branches[1]->d[3];
                asdf->d[5] = asdf->branches[1]->d[5];
                asdf->d[7] = asdf->branches[1]->d[7];
                return;
        case 2: asdf->d[2] = asdf->branches[2]->d[2];
                asdf->d[3] = asdf->branches[2]->d[3];
                asdf->d[6] = asdf->branches[2]->d[6];
                asdf->d[7] = asdf->branches[2]->d[7];
                return;
        case 3: asdf->d[1] = asdf->branches[1]->d[1];
                asdf->d[2] = asdf->branches[2]->d[2];
                asdf->d[3] = asdf->branches[3]->d[3];
                asdf->d[5] = asdf->branches[1]->d[5];
                asdf->d[6] = asdf->branches[2]->d[6];
                asdf->d[7] = asdf->branches[3]->d[7];
                return;
        case 4: asdf->d[4] = asdf->branches[4]->d[4];
                asdf->d[5] = asdf->branches[4]->d[5];
                asdf->d[6] = asdf->branches[4]->d[6];
                asdf->d[7] = asdf->branches[4]->d[7];
                return;
        case 5: asdf->d[1] = asdf->branches[1]->d[1];
                asdf->d[3] = asdf->branches[1]->d[3];
                asdf->d[4] = asdf->branches[4]->d[4];
                asdf->d[5] = asdf->branches[5]->d[5];
                asdf->d[6] = asdf->branches[4]->d[6];
                asdf->d[7] = asdf->branches[5]->d[7];
                return;
        case 6: asdf->d[2] = asdf->branches[2]->d[2];
                asdf->d[3] = asdf->branches[2]->d[3];
                asdf->d[4] = asdf->branches[4]->d[4];
                asdf->d[5] = asdf->branches[4]->d[5];
                asdf->d[6] = asdf->branches[6]->d[6];
                asdf->d[7] = asdf->branches[6]->d[7];
                return;
        case 7: asdf->d[1] = asdf->branches[1]->d[1];
                asdf->d[2] = asdf->branches[2]->d[2];
                asdf->d[3] = asdf->branches[3]->d[3];
                asdf->d[4] = asdf->branches[4]->d[4];
                asdf->d[5] = asdf->branches[5]->d[5];
                asdf->d[6] = asdf->branches[6]->d[6];
                asdf->d[7] = asdf->branches[7]->d[7];
                return;
        default:
            return;
    }
}

void get_xyz_from_children(ASDF* asdf) {

    asdf->X = asdf->branches[0]->X;
    asdf->Y = asdf->branches[0]->Y;
    asdf->Z = asdf->branches[0]->Z;
    
    if (asdf->branches[1])
        asdf->Z.upper = asdf->branches[1]->Z.upper;
    if (asdf->branches[2])
        asdf->Y.upper = asdf->branches[2]->Y.upper;
    if (asdf->branches[4])
        asdf->X.upper = asdf->branches[4]->X.upper;
}


////////////////////////////////////////////////////////////////////////////////

void free_asdf(ASDF* asdf)
{
    if (asdf == NULL)   return;
    
    if (asdf->state == BRANCH)
        for (int i=0; i < 8; ++i)
            free_asdf(asdf->branches[i]);

    free(asdf);
}

////////////////////////////////////////////////////////////////////////////////

int count_cells(const ASDF* const asdf)
{
    if (asdf == NULL) {
        return 0;
    }
        
    int count = 1;
    for (int i=0; i < 8; ++i)
        count += count_cells(asdf->branches[i]);
    
    return count;
}

////////////////////////////////////////////////////////////////////////////////

int get_depth(const ASDF* const asdf)
{
    if (asdf == NULL) {
        return 0;
    }
    
    if (asdf->state == BRANCH) {
        int depth = 1;
        for (int i=0; i < 8; ++i) {
            int d = get_depth(asdf->branches[i]);
            if (d + 1 > depth)
                depth = d+1;
        }
        return depth;
    }
    
    return 1;
}

////////////////////////////////////////////////////////////////////////////////

void find_dimensions(const ASDF* const asdf, int* ni, int* nj, int* nk)
{
    float dx = INFINITY, dy = INFINITY, dz = INFINITY;
    find_min_sizes(asdf, &dx, &dy, &dz);
    
    if (dx > 0) {
        *ni = round((asdf->X.upper - asdf->X.lower)/dx);
    } else {
        *ni = 1;
    }
    
    if (dy > 0) {
        *nj = round((asdf->Y.upper - asdf->Y.lower)/dy);
    } else {
        *nj = 1;
    }

    if (dz > 0) {
        *nk = round((asdf->Z.upper - asdf->Z.lower)/dz);
    } else {
        *nk = 1;
    }
}

void find_min_sizes(const ASDF* const asdf, float* dx, float* dy, float* dz)
{
    if (asdf == NULL)   return;
    
    if (asdf->state == BRANCH) {
        for (int i=0; i < 8; ++i) {
            find_min_sizes(asdf->branches[i], dx, dy, dz);
        }
    } else {
        *dx = fmin(asdf->X.upper - asdf->X.lower, *dx);
        *dy = fmin(asdf->Y.upper - asdf->Y.lower, *dy);
        *dz = fmin(asdf->Z.upper - asdf->Z.lower, *dz);
    }
}

////////////////////////////////////////////////////////////////////////////////

void set_depth(ASDF* const asdf, const int depth)
{
    if (asdf == NULL || asdf->state != BRANCH)  return;
    
    if (depth > 1) {
        for (int a=0; a < 8; ++a) {
            set_depth(asdf->branches[a], depth-1);
        }
    } else {
        asdf->state = LEAF;
    }
}

void clear_depth(ASDF* const asdf)
{
    if (asdf == NULL) return;
    
    if (asdf->state == LEAF) {
        for (int a=0; a < 8; ++a) {
            if (asdf->branches[a]) {
                asdf->state = BRANCH;
                break;
            }
        }
    }
    
    if (asdf->state == BRANCH) {
        for (int a=0; a < 8; ++a) {
            clear_depth(asdf->branches[a]);
        }
    }
}


void make_extrapolation(ASDF* const asdf)
{
    int ni, nj, nk;
    find_dimensions(asdf, &ni, &nj, &nk);
    
    Region r = (Region){.imin = 0,  .jmin = 0,  .kmin = 0,
                        .ni   = ni, .nj   = nj, .nk   = nk,
                        .voxels = ni*nj*nk};

    build_arrays(&r, asdf->X.lower, asdf->Y.lower, asdf->Z.lower,
                 asdf->X.upper - asdf->X.lower,
                 asdf->Y.upper - asdf->Y.lower,
                 asdf->Z.upper - asdf->Z.lower);
    
    int depth = get_depth(asdf);
    for (int d=depth-1; d >= 0; --d) {
    
        Corner* cache = calloc(1, sizeof(Corner));
        cache->value = NAN;
    
        _make_extrapolation(asdf, cache, r, d);
        _apply_extrapolation(asdf, cache, r, d);
        free_corner_cache(cache);
    }
    
    free_arrays(&r);
}


void _make_extrapolation(const ASDF* const asdf, Corner* const cache,
                         const Region r, const int depth)
{
    if (asdf == NULL)   return;
        
    for (int c=0; c < 8; ++c) {
        // Iterate over region corners
        uint16_t i = r.imin + ((c & 4) ? r.ni : 0),
                 j = r.jmin + ((c & 2) ? r.nj : 0),
                 k = r.kmin + ((c & 1) ? r.nk : 0);
        
        // If this corner was already marked as a sample, continue
        Corner* corner = get_corner(cache, i, j, k);
        if (corner->limit == FIXED) continue;

        if (asdf->state == LEAF) {
            corner->value = asdf->d[c];
            corner->limit = FIXED;
        } else if (asdf->state == EMPTY && depth > 0) {
            corner->limit = GEQ_ZERO;
        } else if (asdf->state == FILLED && depth > 0) {
            corner->limit = LT_ZERO;
        }
    }
        
    if (depth > 1) {
        // Get a subcache to speed up cache indexing
        Corner* subcache = corner_subcache(cache,
                                           r.imin, r.imin+r.ni,
                                           r.jmin, r.jmin+r.nj,
                                           r.kmin, r.kmin+r.nk);
    
        if (asdf->state == BRANCH) {
            // Recurse down into the ASDF tree
            Region octants[8];
            octsect_merged(r, asdf, octants);
            for (int a=0; a < 8; ++a) {
                _make_extrapolation(asdf->branches[a], subcache,
                                    octants[a], depth-1);
            }
        } else {
            // Recurse into a dummy ASDF tree (to hit interior
            // points that have been merged together)
            Region octants[8];
            uint8_t bits = octsect(r, octants);
            ASDF dummy = (ASDF){ .state = asdf->state };
            
            // Iterate over all eight possible octants
            for (int a=0; a < 8; ++a) {
            
                if (!(bits & (1 << a))) continue;
            
                // Populate with plausible interior corner values
                for (int c=0; c < 8; ++c) {
                    float x = octants[a].X[(c & 4) ? octants[a].ni : 0],
                          y = octants[a].Y[(c & 2) ? octants[a].nj : 0],
                          z = octants[a].Z[(c & 1) ? octants[a].nk : 0];
                    dummy.d[c] = asdf_interpolate(asdf, x, y, z);
                }
                
                // Continue on down the line with the dummy asdf
                _make_extrapolation(&dummy, subcache,
                                    octants[a], depth-1);
            }
        }
        
    } else if (asdf->state == BRANCH && depth > 0) {
    
        // Boop daboop.
        for (int a=0; a < 8; ++a) {
            Region octants[8];
            octsect_merged(r, asdf, octants);
            for (int a=0; a < 8; ++a) {
                _make_extrapolation(asdf->branches[a], cache,
                                    octants[a], 0);
            }
        }
    
        // We'll now iterate over region corners, picking average values
        // by extrapolating from children nodes.
        for (int c=0; c < 8; ++c) {
        
            // Iterate over region corners
            uint16_t i = r.imin + ((c & 4) ? r.ni : 0),
                     j = r.jmin + ((c & 2) ? r.nj : 0),
                     k = r.kmin + ((c & 1) ? r.nk : 0);
            
            // If this corner was already marked as a sample, continue
            Corner* corner = get_corner(cache, i, j, k);
            if (corner->limit == FIXED)    continue;
            
            float    x = (c & 4) ? asdf->X.upper : asdf->X.lower,
                     y = (c & 2) ? asdf->Y.upper : asdf->Y.lower,
                     z = (c & 1) ? asdf->Z.upper : asdf->Z.lower;
            
            for (int a=0; a < 8; ++a) {
                ASDF* leaf = asdf->branches[a];
                
                if (leaf == NULL || (leaf->state != LEAF &&
                                     leaf->state != BRANCH)) {
                    continue;
                }
                
                if (isnan(corner->value)) {
                    corner->value = asdf_interpolate(leaf, x, y, z);
                } else {
                    float v = asdf_interpolate(leaf, x, y, z);
                    if (v < 0  && corner->limit == GEQ_ZERO) continue;
                    if (v >= 0 && corner->limit == LT_ZERO)  continue;
                    
                    if (fabs(corner->value) > fabs(v)) {
                        corner->value = v;
                    }
                }
                corner->count++;
            }
        }
    }
}

void _apply_extrapolation(ASDF* const asdf, Corner* const cache,
                          const Region r, const int depth)
{
    if (asdf == NULL || asdf->state != BRANCH)   return;
        
    if (depth > 1) {
        
        Region octants[8];
        octsect_merged(r, asdf, octants);
        
        // Get a subcache to speed up cache indexing
        Corner* subcache = corner_subcache(cache,
                                           r.imin, r.imin+r.ni,
                                           r.jmin, r.jmin+r.nj,
                                           r.kmin, r.kmin+r.nk);
        
        // Recurse down into the ASDF tree
        for (int a=0; a < 8; ++a) {
            _apply_extrapolation(asdf->branches[a], subcache,
                                 octants[a], depth-1);
        }
    
    } else {
    
        // We'll now iterate over all 8 corners, applying values
        for (int c=0; c < 8; ++c) {
        
            // Iterate over region corners
            uint16_t i = r.imin + ((c & 4) ? r.ni : 0),
                     j = r.jmin + ((c & 2) ? r.nj : 0),
                     k = r.kmin + ((c & 1) ? r.nk : 0);
    
            Corner* corner = get_corner(cache, i, j, k);
            float val = corner->value, count = 1;
            
            if (corner->limit == FIXED) {
                asdf->d[c] = val;
            } else if (corner->limit == LT_ZERO) {
                asdf->d[c] = fmin(-EPSILON, val/count);
            } else if (corner->limit == GEQ_ZERO) {
                asdf->d[c] = fmax(0, val/count);
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////////

float interpolate(const float d[8], float x, float y, float z,
                  float xmin, float ymin, float zmin,
                  float dx,   float dy,   float dz)
{

    float  xd = dx ? (x - xmin) / dx : 0.5;
    float  yd = dy ? (y - ymin) / dy : 0.5;
    float  zd = dz ? (z - zmin) / dz : 0.5;
    
    float c00 = d[0]*(1-xd) + d[4]*xd;
    float c01 = d[1]*(1-xd) + d[5]*xd;
    float c10 = d[2]*(1-xd) + d[6]*xd;
    float c11 = d[3]*(1-xd) + d[7]*xd;
    
    float  c0 = c00*(1-yd) + c10*yd;
    float  c1 = c01*(1-yd) + c11*yd;
    
    float   c = c0*(1-zd) + c1*zd;

    return c;
}

float asdf_interpolate(const ASDF* const asdf,
                       const float x, const float y, const float z)
{
    return interpolate(asdf->d, x, y, z,
                       asdf->X.lower,
                       asdf->Y.lower,
                       asdf->Z.lower,
                       asdf->X.upper - asdf->X.lower,
                       asdf->Y.upper - asdf->Y.lower,
                       asdf->Z.upper - asdf->Z.lower);
}

Vec3f asdf_gradient(const ASDF* const leaf,
                    const float x, const float y, const float z)
{
    float xmin  = leaf->X.lower,
          ymin  = leaf->Y.lower,
          zmin  = leaf->Z.lower;
          
    float dx    = leaf->X.upper - xmin,
          dy    = leaf->Y.upper - ymin,
          dz    = leaf->Z.upper - zmin;
          
    float xd    = dx ? (x - xmin) / dx : 0.5,
          yd    = dy ? (y - ymin) / dy : 0.5,
          zd    = dz ? (z - zmin) / dz : 0.5;
    
    float c000  = leaf->d[0], c001 = leaf->d[1],
          c010  = leaf->d[2], c011 = leaf->d[3],
          c100  = leaf->d[4], c101 = leaf->d[5],
          c110  = leaf->d[6], c111 = leaf->d[7];
            
    // Numerically find the gradients
    Vec3f G;
    G.x = -(zd - 1)*((yd - 1)*(c000/dx - c100/dx) - (c010/dx - c110/dx)*yd) + (z - zmin)*((yd - 1)*(c001/dx - c101/dx) - (c011/dx - c111/dx)*yd)/dz;
    
    G.y = (zd - 1)*(((xd - 1)*c010 - xd*c110)/dy - ((xd - 1)*c000 - xd*c100)/dy) - (((xd - 1)*c011 - xd*c111)/dy - ((xd - 1)*c001 - xd*c101)/dy)*zd;
    
    G.z = -((yd - 1)*((xd - 1)*c000 - xd*c100) - (y - ymin)*((xd - 1)*c010 - xd*c110)/dy)/dz + ((yd - 1)*((xd - 1)*c001 - xd*c101) - (y - ymin)*((xd - 1)*c011 - xd*c111)/dy)/dz;
    
    return G;
}

/*
# Here is the python code used to find the formulas above
# It should be run in sage, a python CAS environment
var('c000, c001, c010, c011, c100, c101, c110, c111')
var('x, y, z')
var('xmin, ymin, zmin')
var('dx, dy, dz')
xd = (x - xmin) / dx
yd = (y - ymin) / dy
zd = (z - zmin) / dz

c00  = c000*(1-xd) + c100*xd
c01  = c001*(1-xd) + c101*xd
c10  = c010*(1-xd) + c110*xd
c11  = c011*(1-xd) + c111*xd
c0  = c00*(1-yd) + c10*yd
c1  = c01*(1-yd) + c11*yd
c = c0*(1-zd) + c1*zd

print '\n'.join(map(str, c.gradient((x,y,z))))
*/