#ifndef ASDF_H
#define ASDF_H

#include <stdint.h>
#include <stdbool.h>

#include "util/interval.h"
#include "util/region.h"
#include "util/vec3f.h"

struct PackedTree_;
struct Corner_;

////////////////////////////////////////////////////////////////////////////////

enum ASDFstate { FILLED, EMPTY, BRANCH, LEAF };
    
typedef struct ASDF_ {
    enum ASDFstate state;
    Interval X, Y, Z;
    struct ASDF_* branches[8];
    float d[8];
} ASDF;


/*  get_depth
 *
 *  Finds the height of an ASDF.
 */
int get_depth(const ASDF* const asdf);


/*  asdf_root
 *
 *  Generates a root for an ASDF, evaluating the tree at region corners
 *  but without recursing down smaller branches.
 */
ASDF* asdf_root(struct PackedTree_* tree, Region region);


/*  free_asdf
 *
 *  Recursively frees an ASDF data structure.
 */
void free_asdf(ASDF* asdf);


/*  interpolate
 *
 *  Performs general-purpose interpolation.
 *  d should contain samples at all eight corners,
 *  arranged in increasing binary order.
 *      e.g. [xmin, ymin, zmin]  => 000
             [xmin, ymin, zmax]  => 001
             [xmin, ymax, zmin]  => 010
             [xmin, ymax, zmax]  => 011
             etc.
 */
float interpolate(const float d[8], float x, float y, float z,
                  float xmin, float ymin, float zmin,
                  float dx,   float dy,   float dz);


/*  asdf_interpolate
 *
 *  Performs interpolation within an ASDF cell.
 */
float asdf_interpolate(const ASDF* const asdf, const float x,
                       const float y, const float z);
                       
                       
/*  asdf_gradient
 *
 *  Finds the gradient at a point inside an ASDF cell.
 */
Vec3f asdf_gradient(const ASDF* const leaf, const float x,
                    const float y, const float z);


/*  simplify
 *
 *  Simplifies an ASDF cell by merging children (non-recursive)
 */
void simplify(ASDF* asdf);


/*  merge_cells
 *
 *  Enlarges A to include B, copying over appropriate d[i] values
 *  Axis is 4, 2, or 1, which represent x, y, or z axis merging.
 */
void merge_cells(ASDF* A, ASDF* B, int axis);


/*  find_dimensions
 *
 *  Finds the number of voxels along each dimension.
 */
void find_dimensions(const ASDF* const asdf, int* ni, int* nj, int* nk);


/*  find_min_sizes
 *
 *  Finds the minimum cell sizes along each dimension.
 */
void find_min_sizes(const ASDF* const asdf, float* dx, float* dy, float* dz);


/*  count_cells
 *
 *  Returns how much cells exist in this ASDF
 */
int count_cells(const ASDF* const asdf);


void get_d_from_children(ASDF* asdf);
void get_xyz_from_children(ASDF* asdf);


/*  make_extrapolation
 *
 *  Populates branch cells with d values that do a good job
 *  reconstructing their children cells.
 */
void make_extrapolation(ASDF* const asdf);
void _make_extrapolation(const ASDF* const asdf,
                         struct Corner_* const cache,
                         const Region r, const int depth);


/*  _apply_extrapolation
 *
 *  Applies the average values found in _make_extrapolation to corner d values.
 */
void _apply_extrapolation(ASDF* const asdf,
                          struct Corner_* const cache,
                          const Region r, const int depth);
                          

ASDF*  build_asdf_region(struct PackedTree_* tree, Region region);
ASDF* _build_asdf_region(float* result, Region region, int kstride, int jstride);

////////////////////////////////////////////////////////////////////////////////

#endif