#ifndef CACHE_H
#define CACHE_H

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

#include "util/region.h"
#include "util/vec3f.h"

struct ASDF_;

/*  Corner
 *
 *  This structure implements a tree-based cache indexed by i, j, k
 *  position (as 16-bit unsigned integers) in a discrete lattice.
 *
 *  Each cache has a depth value.  For any given i, j, k value, the
 *  corresponding cache corner is reached when bits below the depth-th
 *  bit (counting from the top) are all zeros.
 *
 *  For example (showing only the top eight bits)
 *      i = 0b11010000, j = 0b10111000, k = 0b00000000
 *  will be at a depth of 5 (j is the limiting factor)
 */
typedef struct Corner_ {
    uint8_t depth;
    float   value;
    int     count;
    enum { NONE, LT_ZERO, GEQ_ZERO, FIXED } limit;
    struct Corner_* branches[8];
} Corner;


/*  free_cache
 *
 *  Recursively frees a corner cache.
 */
void free_corner_cache(Corner* const cache);


/*  write_cache
 *
 *  Writes a cache out to a file, discretizing to 16 bits centered on zero
 */
void  write_cache(const Corner* const cache, FILE* file);
void _write_cache(const Corner* const cache, FILE* file,
                  const float scale);


/*  read_cache
 *
 *  Reads a cache in from a file, with 16 bits of range centered on zero
 */
Corner*  read_cache(FILE* file);
Corner* _read_cache(FILE* file, const int depth,
                    const float scale);


/*  get_range
 *
 *  Returns the minimum and maximum value found in the cache.
 */
void get_range(const Corner* const cache, float* const min, float* const max);


/*  get_corner
 *
 *  Finds the corner corresponding to the given i, j, k position.
 */
Corner* get_corner(Corner* const cache,
                   const uint16_t i,
                   const uint16_t j,
                   const uint16_t k);


/*  corner_subcache
 *
 *  Returns a subset of the cache based on the similar
 *  upper bits of X, Y, Z.
 */
Corner* corner_subcache(Corner* const cache,
                        const uint16_t imin, const uint16_t imax,
                        const uint16_t jmin, const uint16_t jmax,
                        const uint16_t kmin, const uint16_t kmax);
              
              
/*  fill_corner_cache
 *
 *  Creates a populated cache with corner values of ASDF cells.
 */
Corner* fill_corner_cache(const struct ASDF_* const asdf);
void   _fill_corner_cache(const struct ASDF_* const asdf,
                   Corner* const cache, const Region r);

////////////////////////////////////////////////////////////////////////////////

typedef struct Edge_ {
    uint8_t depth;
    Vec3f   pos;
    Vec3f   gradient;
    _Bool   valid;
    uint32_t id;
    struct Edge_* branches[64];
} Edge;

Edge* get_edge(Edge* const cache,
               const uint16_t i1, const uint16_t j1, const uint16_t k1,
               const uint16_t i2, const uint16_t j2, const uint16_t k2);

Edge* edge_subcache(Edge* const cache,
                    const uint16_t imin, const uint16_t imax,
                    const uint16_t jmin, const uint16_t jmax,
                    const uint16_t kmin, const uint16_t kmax);

void free_edge_cache(Edge* cache);

uint32_t cached_vertex_count(Edge* const cache);

#endif