#include <stdio.h>
#include <math.h>

#include "asdf/asdf.h"
#include "asdf/render.h"

#include "util/switches.h"
#include "util/macros.h"
#include "util/vec3f.h"

////////////////////////////////////////////////////////////////////////////////

void render_asdf(const ASDF const* asdf, const Region r_,
                 const float M[4], uint16_t*const*const depth,
                 uint16_t*const*const shaded, uint8_t (**normals)[3])
{
    if (!asdf || asdf->state == EMPTY)  return;

    Region r = rot_bound_region(asdf, r_, M);
    
    if (r.voxels == 0)  return;
    
    else if (asdf->state == FILLED) {
        for (int j=0; j < r.nj; ++j) {
            for (int i=0; i < r.ni; ++i) {
                if (r.L[r.nk] <= depth[r.jmin + j][r.imin + i])
                    continue;
                
                for (int k=r.nk-1; k >= 0; --k) {
                    uint16_t L = r.L[k];

                    if (L < depth[r.jmin+j][r.imin+i])
                        break;
                    
                    // Transform a pixel-sized cube into the cell's coordinate
                    // frame and make sure there's some overlap.
                    Interval pX, pY, pZ;
                    deproject_cube( (Interval){r.X[i], r.X[i+1]}, 
                                    (Interval){r.Y[j], r.Y[j+1]},
                                    (Interval){r.Z[k], r.Z[k+1]},
                                    &pX, &pY, &pZ, M );
                    
                    // If there's no overlap between the voxel and
                    // the pixel (in the voxel's coordinate frame),
                    // then skip this pixel
                    if (pX.upper < asdf->X.lower || asdf->X.upper < pX.lower ||
                        pY.upper < asdf->Y.lower || asdf->Y.upper < pY.lower ||
                        pZ.upper < asdf->Z.lower || asdf->Z.upper < pZ.lower)
                    {
                        continue;
                    }

                    depth[r.jmin+j][r.imin+i] = L;
                    break;
                }
            }
        }
    }
    
    // For a leaf cell or pixel-wide cell, render as a leaf
    else if (asdf->state == LEAF) {
        render_leaf(asdf, r, M, depth, shaded, normals);
    }
    
    // From a branching cell, recurse down all branches
    else if (asdf->state == BRANCH) {
        for (int i=0; i < 8; ++i) {
            render_asdf(asdf->branches[i], r, M,
                              depth, shaded, normals);
        }
    }

}

////////////////////////////////////////////////////////////////////////////////

void render_leaf(const ASDF* const leaf, const Region r,
                 const float M[4], uint16_t*const*const depth,
                 uint16_t*const*const shaded, uint8_t (**normals)[3])
{

    // Loop over X, Y, and Z points
    for (int j=0; j < r.nj; ++j) {
        for (int i=0; i < r.ni; ++i) {

            // If we can't brighten this point, then return early
            if (r.L[r.nk] <= depth[r.jmin + j][r.imin + i])
                continue;
            
            // Loop downwards over Z values
            for (int k=r.nk-1; k >= 0; --k) {
            
                // If we can't brighten the point, return early
                uint16_t L = r.L[k];
                if (L < depth[r.jmin+j][r.imin+i])  break;
                                    
                // Project a pixel-sized cube into the cell's frame.
                Interval pX, pY, pZ;

                deproject_cube( (Interval){r.X[i], r.X[i+1]}, 
                                (Interval){r.Y[j], r.Y[j+1]},
                                (Interval){r.Z[k], r.Z[k+1]},
                                &pX, &pY, &pZ, M );
                
                // If there's no overlap, then skip this pixel
                if (pX.upper < leaf->X.lower || leaf->X.upper < pX.lower ||
                    pY.upper < leaf->Y.lower || leaf->Y.upper < pY.lower ||
                    pZ.upper < leaf->Z.lower || leaf->Z.upper < pZ.lower)
                {
                    continue;
                }

                // Now, check each of the pixel's corners to decide
                // if we should color it in.
                Vec3f p;
                float c = 1;
                for (int i=0; c >= 0 && i < 8; ++i) {
                    p = (Vec3f) {(i & 4) ? pX.upper : pX.lower,
                                 (i & 2) ? pY.upper : pY.lower,
                                 (i & 1) ? pZ.upper : pZ.lower};
                    c = asdf_interpolate(leaf, p.x, p.y, p.z);
                }

                // If we've found an intersection, then color it in!
                if (c < 0) {
                
                
                    // Save the depth
                    depth[r.jmin+j][r.imin+i] = L;
                    
                    // Find the gradient
                    Vec3f gradient = asdf_gradient(leaf, p.x, p.y, p.z);
                    Vec3f normal = normalize(project(gradient, M));

                    // If this point has a front-facing normal, shade it.
                    if (normal.z >= 0) {
                        float xy = sqrt(pow(normal.x, 2) + pow(normal.y, 2));
                        float alpha = fabs(atan2(normal.z, xy) - M_PI_2);
                        
                        float shade = 65535 * sqrt(1 - alpha / M_PI_2);
                        shaded[r.jmin+j][r.imin+i] = shade;
                    }
                    
                    // And draw colors to show the normal
                    normals[r.jmin+j][r.imin+i][0] = 255*(normal.x/2+0.5);
                    normals[r.jmin+j][r.imin+i][1] = 255*(normal.y/2+0.5);
                    normals[r.jmin+j][r.imin+i][2] = 255*(normal.z/2+0.5);
                    
                    // We can break here, since we'll only be going down
                    // in z value, so later points will be culled.
                    break;
                }
                
            } // End of Z loop
        }   // End of X loop
    }   // End of Y loop
}

////////////////////////////////////////////////////////////////////////////////

void draw_asdf_cells(ASDF* a, Region r, uint8_t (**img)[3])
{
    if (!a)  return;

    uint8_t R = 0, G = 0, B = 0;
    if (a->state == BRANCH) {
        for (int i=0; i < 8; i++)
            draw_asdf_cells(a->branches[i], r, img);
        return;
    } else if (a->state == EMPTY) {
        B = 255;
        R = 50;
        G = 100;
    } else if (a-> state == FILLED)
        R = 255;
    else if (a->state == LEAF) {
        G = 255;
    }

    int imin, imax, jmin, jmax;

    if (a->X.upper <= r.X[0] || a->X.lower >= r.X[r.ni] ||
        a->Y.upper <= r.Y[0] || a->Y.lower >= r.Y[r.nj])
        return;

    float dx = r.X[r.ni] - r.X[0];
    imin = BOUND(round(r.ni*(a->X.lower - r.X[0]) / dx), 0, r.ni);
    imax = BOUND(round(r.ni*(a->X.upper - r.X[0]) / dx), 1, r.ni);

    float dy = r.Y[r.nj] - r.Y[0];
    jmin = BOUND(round(r.nj*(a->Y.lower - r.Y[0]) / dy), 0, r.nj);
    jmax = BOUND(round(r.nj*(a->Y.upper - r.Y[0]) / dy), 1, r.nj);

    for (int i=imin; i < imax; ++i) {
        img[jmin][i][0] = R;
        img[jmin][i][1] = G;
        img[jmin][i][2] = B;
        img[jmax-1][i][0] = R;
        img[jmax-1][i][1] = G;
        img[jmax-1][i][2] = B;
    }
    for (int j=jmin; j < jmax; ++j) {
        img[j][imin][0] = R;
        img[j][imin][1] = G;
        img[j][imin][2] = B;
        img[j][imax-1][0] = R;
        img[j][imax-1][1] = G;
        img[j][imax-1][2] = B;
    }

}

