#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "node/node.h"
#include "node/printers.h"
#include "math/math_f.h"

#include "util/switches.h"

// Non-recursively clone a node, linking to cloned children if
// non-constant or original children if constant.
Node* clone_node(Node* n)
{       
    // Allocate memory and copy everything over
    Node* clone = malloc(sizeof(Node));
    memcpy(clone, n, sizeof(Node));
    
    // Link to cloned children, and pick the greater of their weights
    // (normally won't change the node's rank, except when doing things
    //  like a map operator, where we hack the clone address to point to
    //  a remapped node).
    int lhs_rank = -1, rhs_rank = -1;
    if (n->lhs) {
        clone->lhs = n->lhs->clone_address;
        lhs_rank   = clone->lhs->rank;
    }
    if (n->rhs) {
        clone->rhs = n->rhs->clone_address;
        rhs_rank   = clone->rhs->rank;
    }
    clone->rank = 1 + ((lhs_rank > rhs_rank) ? lhs_rank : rhs_rank);
    
    // Record the address of the new clone, so that clones of
    // its parents can be adjusted to point in the right place
    n->clone_address = clone;
    
    return clone;
}

////////////////////////////////////////////////////////////////////////////////

Node* binary_n(Node* lhs, Node* rhs)
{
    Node* n = malloc(sizeof(Node));
    
    _Bool constant = lhs->constant && rhs->constant;
    
    *n = (Node) {
        .constant   = constant,
        .rank       = constant ? 0 : 1 + (lhs->rank > rhs->rank ?
                                          lhs->rank : rhs->rank),
        .marked     = false,
        .lhs        = lhs,
        .rhs        = rhs,
        .clone_address = NULL,
    };

    return n;
} 

// Macro to define a binary node with a name, function, and opcode
#define BINARY_NODE(N, F, O)                                \
Node* N(Node* lhs, Node* rhs)                               \
{                                                           \
    Node* n = binary_n(lhs, rhs);                           \
    n->opcode = O;                                          \
    if (n->constant) {                                      \
        fill_results(n, F(n->lhs->results.f,                \
                          n->rhs->results.f));              \
        n->opcode = OP_CONST;                               \
        n->lhs = NULL;                                      \
        n->rhs = NULL;                                      \
    }                                                       \
    return n;                                               \
}

BINARY_NODE(add_n, add_f, OP_ADD)
BINARY_NODE(sub_n, sub_f, OP_SUB)
BINARY_NODE(mul_n, mul_f, OP_MUL)
BINARY_NODE(div_n, div_f, OP_DIV)

BINARY_NODE(min_n, min_f, OP_MIN)
BINARY_NODE(max_n, max_f, OP_MAX)
BINARY_NODE(pow_n, pow_f, OP_POW)

    
////////////////////////////////////////////////////////////////////////////////

Node* unary_n(Node* arg)
{
    Node* n = malloc(sizeof(Node));
    
    _Bool constant = arg->constant;
    
    *n = (Node) {
        .constant   = constant,
        .rank       = constant ? 0 : 1 + arg->rank,
        .marked     = false,
        .lhs        = arg,
        .rhs        = NULL,
        .clone_address = NULL,
    };
    
    return n;
}

#define UNARY_NODE(N, F, O)                                 \
Node* N(Node* child)                                        \
{                                                           \
    Node* n = unary_n(child);                               \
    n->opcode = O;                                          \
    if (n->constant) {                                      \
        fill_results(n, F(n->lhs->results.f));              \
        n->opcode = OP_CONST;                               \
        n->lhs = NULL;                                      \
        n->rhs = NULL;                                      \
    }                                                       \
    return n;                                               \
}

UNARY_NODE(square_n, square_f, OP_SQUARE)
UNARY_NODE(sqrt_n, sqrt_f, OP_SQRT)
UNARY_NODE(sin_n, sin_f, OP_SIN)
UNARY_NODE(cos_n, cos_f, OP_COS)
UNARY_NODE(tan_n, tan_f, OP_TAN)
UNARY_NODE(neg_n, neg_f, OP_NEG)
UNARY_NODE(buf_n, buf_f, OP_BUF)

////////////////////////////////////////////////////////////////////////////////

Node* nonary_n(void)
{
    Node* n = malloc(sizeof(Node));
    
    *n = (Node) {
        .constant   = false,
        .rank       = 0,
        .marked     = false,
        .lhs        = NULL,
        .rhs        = NULL,
        .clone_address = NULL,
    };

    return n;
}

Node* constant_n(float value)
{
    Node* n = nonary_n();
    
    n->constant = true;
    n->opcode   = OP_CONST;
    
    fill_results(n, value);
    return n;
}

Node* X_n()
{
    Node* n = nonary_n();
    n->opcode = OP_X;
    return n;
}

Node* Y_n()
{
    Node* n = nonary_n();
    n->opcode = OP_Y;
    return n;
}

Node* Z_n()
{
    Node* n = nonary_n();
    n->opcode = OP_Z;
    return n;
}