#ifndef TREE_H
#define TREE_H

#include "node/opcodes.h"

/*  MathTree
 *
 *  A structure containing nodes organized by rank and opcode.
 */
typedef struct MathTree_ {

    // 3D array  indexed by [level][opcode][node]
    struct Node_** ((*nodes)[LAST_OP]);

    // 2D arrays indexed by [level][opcode]
    unsigned       (*active)[LAST_OP];

    // 1D array of node pointers
    struct Node_ **constants;
    
    // Size of the constants array
    unsigned       num_constants;
    
    // Pointer to tree head
    struct Node_* head;
    
    // Number of levels in the tree
    unsigned num_levels;
} MathTree;


/*  free_tree
 *
 *  Frees a tree and associated nodes.
 */
void free_tree(MathTree* tree);


/*  deduplicate_tree
 *
 *  Removes duplicate nodes from a well-formed MathTree
 */
void deduplicate_tree(MathTree* tree);


/*  deduplicate_constants
 *
 *  Deduplicates a list of constant nodes, returning the number
 *  of nodes shifted to the back of the list.  It is up to the
 *  caller to free the extra nodes and reallocate the array's
 *  space.
 */
int deduplicate_constants(struct Node_** constants, unsigned num);


/*  deduplicate_row
 *
 *  Removes duplicates from a row, returning the number of nodes
 *  shifted to the back of the list.  Also adjusts pointers if
 *  clone_address of children is populated.  As above, the caller
 *  is responsible for freeing the extra nodes and reallocating the
 *  array.
 */
int deduplicate_row(struct Node_** nodes, unsigned num);


/*  print_tree
 *
 *  Prints a math tree to stdout.
 */
void print_tree(MathTree* tree);
void print_tree_verbose(MathTree* tree);


/*  fprintf_tree
 *
 *  Prints a math tree to a given file descriptor.
 */
void fprint_tree(MathTree* tree, int fd);
void fprint_tree_verbose(MathTree* tree, int fd);
 
/*  map_tree
 *
 *  Replaces X, Y, Z with new expressions.
 */
MathTree* map_tree(MathTree* target, MathTree* X, MathTree* Y, MathTree* Z);


/*  count_nodes
 *
 *  Counts nodes, storing them in the array 'count' indexed as
 *  count[rank][opcode].  'count' must have a size equal to the
 *  largest rank among nodes in the tree.
 */
void count_nodes(MathTree* tree, unsigned (*count)[LAST_OP]);


/*  rejigger_tree
 *
 *  Modifies a tree so that all nodes are in the appropriate
 *  rank level.  This may include reallocating the arrays of
 *  node pointers and active node counts.
 */
void rejigger_tree(MathTree* tree);


/*  clone_tree
 *
 *  Clones an entire tree.
 */
MathTree* clone_tree(MathTree* orig);

/*  mark
 *
 *  Sets the mark variable to true on all nodes.
 */
void  mark(MathTree* tree);


/*  unmark
 *
 *  Sets the mark variable to false on all nodes.
 */
void  unmark(MathTree* tree);


////////////////////////////////////////////////////////////////////////////////

// Binary operations
MathTree* add_tree(MathTree* left, MathTree* right);
MathTree* sub_tree(MathTree* left, MathTree* right);
MathTree* mul_tree(MathTree* left, MathTree* right);
MathTree* div_tree(MathTree* left, MathTree* right);

MathTree* min_tree(MathTree* left, MathTree* right);
MathTree* max_tree(MathTree* left, MathTree* right);
MathTree* pow_tree(MathTree* left, MathTree* right);

// Unary arithmetic operators
MathTree* square_tree(MathTree* n);
MathTree* sqrt_tree(MathTree* n);
MathTree* neg_tree(MathTree* n);
//MathTree* sin_tree(MathTree* n);
//MathTree* cos_tree(MathTree* n);
//MathTree* tan_tree(MathTree* n);

// Constants
MathTree* constant_tree(float value);

// Variables
MathTree* X_tree(void);
MathTree* Y_tree(void);
MathTree* Z_tree(void);

#endif